#include <GL/gl.h>
#include <GL/glu.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "timer.h"
#include "wiimote.h"
#include <ode/ode.h>
#include "box.h"

#define PARTICLE_NUM 35

typedef struct {
  dReal x, y, z;
} Point3d;

typedef struct {
  int shape;
  int level;
  Point3d position;
  int active;
  dBodyID body;
  dGeomID geom;
  dMass m;
} Particle;

static Particle particles[PARTICLE_NUM];
static dWorldID world;
static dSpaceID space;
static dJointGroupID contactgroup;

static GLfloat zoom = -50;
static GLfloat tilt = 0;
static GLfloat spin;

static GLuint title_texture;
static GLuint star_texture;
static GLuint loading_texture;
static GLuint wiimote_loading_texture;
static GLuint credits_texture;

/* Image type - contains height, width, and data */
struct Image {
    unsigned long sizeX;
    unsigned long sizeY;
    char *data;
};
typedef struct Image Image;

/*
 * getint and getshort are help functions to load the bitmap byte by byte on 
 * SPARC platform (actually, just makes the thing work on platforms of either
 * endianness, not just Intel's little endian)
 */

static unsigned int getint(fp)
     FILE *fp;
{
  int c, c1, c2, c3;

  // get 4 bytes
  c = getc(fp);  
  c1 = getc(fp);  
  c2 = getc(fp);  
  c3 = getc(fp);
  
  return ((unsigned int) c) +   
    (((unsigned int) c1) << 8) + 
    (((unsigned int) c2) << 16) +
    (((unsigned int) c3) << 24);
}

static unsigned int getshort(fp)
     FILE *fp;
{
  int c, c1;
  
  //get 2 bytes
  c = getc(fp);  
  c1 = getc(fp);

  return ((unsigned int) c) + (((unsigned int) c1) << 8);
}

// quick and dirty bitmap loader...for 24 bit bitmaps with 1 plane only.  
// See http://www.dcs.ed.ac.uk/~mxr/gfx/2d/BMP.txt for more info.

static int ImageLoad(char *filename, Image *image) 
{
    FILE *file;
    unsigned long size;                 // size of the image in bytes.
    unsigned long i;                    // standard counter.
    unsigned short int planes;          // number of planes in image (must be 1) 
    unsigned short int bpp;             // number of bits per pixel (must be 24)
    char temp;                          // used to convert bgr to rgb color.

    // make sure the file is there.
    if ((file = fopen(filename, "rb"))==NULL) {
      printf("File Not Found : %s\n",filename);
      return 0;
    }
    
    // seek through the bmp header, up to the width/height:
    fseek(file, 18, SEEK_CUR);

    // No 100% errorchecking anymore!!!

    // read the width
    image->sizeX = getint (file);
    //printf("Width of %s: %lu\n", filename, image->sizeX);
    
    // read the height 
    image->sizeY = getint (file);
    //printf("Height of %s: %lu\n", filename, image->sizeY);
    
    // calculate the size (assuming 24 bits or 3 bytes per pixel).
    size = image->sizeX * image->sizeY * 3;

    // read the planes
    planes = getshort(file);
    if (planes != 1) {
	printf("Planes from %s is not 1: %u\n", filename, planes);
	return 0;
    }

    // read the bpp
    bpp = getshort(file);
    if (bpp != 24) {
      printf("Bpp from %s is not 24: %u\n", filename, bpp);
      return 0;
    }
	
    // seek past the rest of the bitmap header.
    fseek(file, 24, SEEK_CUR);

    // read the data. 
    image->data = (char *) malloc(size);
    if (image->data == NULL) {
	printf("Error allocating memory for color-corrected image data");
	return 0;	
    }

    if ((i = fread(image->data, size, 1, file)) != 1) {
	printf("Error reading image data from %s.\n", filename);
	return 0;
    }

    for (i=0;i<size;i+=3) { // reverse all of the colors. (bgr -> rgb)
	temp = image->data[i];
	image->data[i] = image->data[i+2];
	image->data[i+2] = temp;
    }

    // we're done.
    return 1;
}

// Load Bitmaps And Convert To Textures
static GLuint LoadGLTexture(char *filename) 
{	
    // Load Texture
    Image *image1;
    GLuint texture;
    
    // allocate space for texture
    image1 = (Image *) malloc(sizeof(Image));
    if (image1 == NULL) {
	printf("Error allocating space for image");
	exit(0);
    }

    if (!ImageLoad(filename, image1)) {
	exit(1);
    }        

    // Create Textures	
    glGenTextures(1, &texture);

    // linear filtered texture
    glBindTexture(GL_TEXTURE_2D, texture);   // 2d texture (x and y size)
    
    glTexImage2D(GL_TEXTURE_2D, 0, 3, image1->sizeX, image1->sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, image1->data);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR );
    gluBuild2DMipmaps(GL_TEXTURE_2D, 3, image1->sizeX, image1->sizeY, GL_BGR, GL_UNSIGNED_BYTE, image1->data);
    /*
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR); // scale linearly when image bigger than texture
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR); // scale linearly when image smalled than texture
    */
    return texture;
};

static float rand1()
{
  return ((rand() / (RAND_MAX + 1.0)));
}

// this is called by dSpaceCollide when two objects in space are
// potentially colliding.
static void nearCallback (void *data, dGeomID o1, dGeomID o2)
{
 dBodyID b1 = dGeomGetBody(o1);
 dBodyID b2 = dGeomGetBody(o2);
 dContact contact;  
 contact.surface.mode = dContactBounce; // | dContactSoftCFM;
 // friction parameter
 //contact.surface.mu = dInfinity;
 // bounce is the amount of "bouncyness".
 contact.surface.bounce = 1.0;
 // bounce_vel is the minimum incoming velocity to cause a bounce
 contact.surface.bounce_vel = 0;//0.1;
 // constraint force mixing parameter
 contact.surface.soft_cfm = 0.001;
 if (dCollide (o1,o2,1,&contact.geom,sizeof(dContact))) {
   dJointID c = dJointCreateContact (world,contactgroup,&contact);
   dJointAttach (c,b1,b2);
 }
}

void init_particle(Particle *p, int shape, int level)
{
  p->body = dBodyCreate(world);
  p->geom = dCreateSphere(space, 1.0);
  dGeomDisable(p->geom);
  dMassSetSphere (&p->m, 1, 0.2);
  dBodySetMass(p->body, &p->m);
  dGeomSetBody(p->geom, p->body);
  
  p->shape = shape;
  p->level = level;
  p->active = 0;
}

void throw_particle(int index)
{
  Particle *p = &particles[index];
  //printf("throwing particle of shape %d at level %d\n", shape, level);
  
  dBodySetPosition (p->body, 0, 0, 0);
  dGeomEnable(p->geom);
  dBodyAddForce(p->body, 20, 24, 13);
  p->active = 1;
}

void throw_particle_intro()
{
  Particle *p = &particles[0];
  //printf("throwing particle of shape %d at level %d\n", shape, level);
  
  dBodySetPosition (p->body, -2.5f, -2.05, 0);
  dGeomEnable(p->geom);
  dBodyAddForce(p->body, 0, 0, 153);
  p->active = 1;
  
  draw_box();
}

void kill_particle(int index)
{
  Particle *p = &particles[index];
  dGeomDisable(p->geom);
  
  p->active = 0;
}

void kill_particle_intro()
{
  kill_particle(0);
}

void init_world()
{
  int i, j;
  Particle *p;
  //dInitODE ();
  dInitODE2(0);

  world = dWorldCreate();
  space = dHashSpaceCreate (0);
  dCreatePlane (space,0,1,0,-10);
  dCreatePlane (space,0,-1,0,-10);
  dCreatePlane (space,-1,0,0,-10);
  dCreatePlane (space,1,0,0,-10);
  dCreatePlane (space,0,0,-1,-30);
  dCreatePlane (space,0,0,1,-5);
  contactgroup = dJointGroupCreate (0);
  dWorldSetGravity(world, 0, 0, 0);

  for (i=0; i<5; i++)
    for (j=0; j<7; j++)
    {
      //printf("init particle %d %d %d\n", get_groovy_index(i, j), i, j);
      p = &particles[get_groovy_index(i, j)];
      init_particle(p, i, j);
    }
}

void update_world()
{
  const dReal *pos;
  const dReal *R;
  double simstep = 0.01; // s simulation steps
  double dt = frame_ticks / 1000.0;
  int nrofsteps = (int) ceilf(dt/simstep);
  int i;
  Particle *p;
  
  //printf("dt: %f, n steps: %d\n", dt, nrofsteps);

  for (i=0; i<nrofsteps; i++)
  {
    dSpaceCollide (space, 0, &nearCallback);
    dWorldQuickStep (world, simstep);
    dJointGroupEmpty (contactgroup);
  }

  for (i=0; i<PARTICLE_NUM; i++)
  {
    p = &particles[i];
    //printf("update particule %d\n", i);
    pos = dGeomGetPosition(p->geom);
    R = dGeomGetRotation(p->geom);
    p->position.x = pos[0];
    p->position.y = pos[1];
    p->position.z = pos[2];
  }
  //printf("%f\t%f\t%f\n", particles[0].position.x, particles[0].position.y, particles[0].position.z);
}

void destroy_world()
{
}

void init_box()
{
  init_world();
  star_texture = LoadGLTexture("star.bmp");
  loading_texture = LoadGLTexture("img/loading.bmp");
  wiimote_loading_texture = LoadGLTexture("img/wiimote.bmp");
  title_texture = LoadGLTexture("img/title.bmp");
  credits_texture = LoadGLTexture("img/end.bmp");
  glEnable(GL_TEXTURE_2D);                    // Enable texture mapping.

  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);	// This Will Clear The Background Color To Black
  glClearDepth(1.0);				// Enables Clearing Of The Depth Buffer

  glShadeModel(GL_SMOOTH);			// Enables Smooth Color Shading
  
  /* setup blending */
  glBlendFunc(GL_SRC_ALPHA,GL_ONE);	        // Set The Blending Function For Translucency
  glEnable(GL_BLEND);                         // Enable Blending

  glDisable(GL_DEPTH_TEST);
}

void draw_box()
{
  static float hit_acc = 0;
  float seconds;
  float speed = 0.2;
  static float total_ticks = 0;
  int r, g, b;
  Particle *p;
  int i;
  
  seconds = frame_ticks / 1000.0;
  total_ticks += frame_ticks;

  hit_acc += seconds;
  
  update_world();  
  
/*
  dCreatePlane (space,0,1,0,-10);
  dCreatePlane (space,0,-1,0,-10);
  dCreatePlane (space,-1,0,0,-10);
  dCreatePlane (space,1,0,0,-10);
  dCreatePlane (space,0,0,-1,-30);
  dCreatePlane (space,0,0,1,-5);
*/  
/*  
  glLoadIdentity();
  glTranslatef(0, 0, -50);
    r = 0;
    g = 0;
    b = 255;
    glColor4ub(r, g, b, 255);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,-1.0f, 0.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,-1.0f, 0.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f, 1.0f, 0.0f);
    glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f, 1.0f, 0.0f);
    glEnd();
*/
  glBindTexture(GL_TEXTURE_2D, star_texture);

  for (i=0; i<PARTICLE_NUM; i++)
  {
    p = &particles[i];
    if (p->active)
    {
      glLoadIdentity();
      glTranslatef(0.0f, 0.0f, zoom);
      
      glTranslatef(p->position.x, p->position.y, p->position.z);
      glScalef(5, 5, 5);
      
      switch (p->shape)
      {
        case 0:
          r = 255 - p->level / 7.0 * 100;
          g = p->level / 7.0 * 100;
          b = 0;
          break;
        
        case 1:
          r = p->level / 7.0 * 100;
          g = 255 - p->level / 7.0 * 100 ;
          b = 0;
          break;
        
        case 2:
          r = p->level / 7.0 * 100;
          g = 255 - p->level / 7.0 * 100 ;
          b = p->level / 7.0 * 100 + 100;
          break;
        
        case 3:
          r = 0;
          g = 255 - p->level / 7.0 * 100 ;
          b = p->level / 7.0 * 10 + 50;
          break;
        
        case 4:
          r = 0;
          g = p->level / 7.0 * 100 + 50;
          b = p->level / 7.0 * 100 + 100;
          break;
      }
      
      glColor4ub(r, g, b, 255);
      glBegin(GL_QUADS);
      glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,-1.0f, 0.0f);
      glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,-1.0f, 0.0f);
      glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f, 1.0f, 0.0f);
      glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f, 1.0f, 0.0f);
      glEnd();
    }
  }
}

static void start_2d_display()
{
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0,1,1,0,-1,1);
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
}

static void stop_2d_display()
{
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
}

static void draw_bmp(GLuint bmp, float x, float y, float size, float ratio, float angle)
{
  start_2d_display();
  //float screen_ratio = (float)screen_width / (float)screen_height;
  glBindTexture( GL_TEXTURE_2D, bmp );
  
  glTranslatef(x,y,0);
  /*
  if (screen_ratio > ratio)
    glScalef(size * ratio / screen_ratio, size, 1.0);
  else if (screen_ratio < ratio)
    glScalef(size, size * screen_ratio / ratio, 1.0);
  */
  glScalef(size, size / ratio, 1.0);
  
  glRotatef(180,1,0,0);
  glRotatef(angle,0,0,1);
  glBegin(GL_TRIANGLE_STRIP);
  glTexCoord2d(0, 0);
  glVertex3f(-0.5, -0.5, 0);
  glTexCoord2d(0, 1);
  glVertex3f(-0.5, 0.5, 0);
  glTexCoord2d(1, 0);
  glVertex3f(0.5, -0.5, 0);
  glTexCoord2d(1, 1);
  glVertex3f(0.5, 0.5, 0);
  glEnd();
  stop_2d_display();
}

void draw_title()
{
  glColor4ub(255, 255, 255, 255);
  draw_bmp(title_texture, 0.7, 0.35, 0.3, 446.0/99.0, 0);
}

void draw_loading_wiimote()
{
  glColor4ub(232, 209, 209, 255);
  draw_bmp(wiimote_loading_texture, 0.7, 0.45, 0.3, 362.0/77.0, 0);
}

void draw_loading()
{
  glColor4ub(232, 209, 209, 255);
  draw_bmp(loading_texture, 0.7, 0.5, 0.2, 204.0/39.0, 0);
}

void draw_credits()
{
  glColor4ub(232, 209, 209, 255);
  draw_bmp(credits_texture, 0.5, 0.5, 1.0, 1024.0/768.0, 0);
}

void shutdown_box()
{
  destroy_world();
}
